# -*- coding: utf-8 -*-
"""
This module contains Service implementation of BeatPort.
http://beatport.com
"""


from . import Service, HTMLMixin

from six import text_type


##############################################################################


class BeatPort(HTMLMixin, Service):
    """
    Implementation of Service which is intended to parse BeatPort.
    """

    def fetch_tracks(self, soup):
        return soup.select("table.track-grid tr.track-grid-content")

    def fetch_name(self, soup, container):
        column = container.find("td", class_="titleColumn")
        elements = column.find("a", class_="txt-larger").find_all("span")
        return " ".join(text_type(elem) for elem in elements)

    def fetch_track_length(self, soup, container):
        element = container.find_all("td")[-2].span
        return text_type(element).split("/")[0].strip()



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of RateYourMusic.
http://rateyourmusic.com
"""


from . import SchemaOrgService


##############################################################################


class RateYourMusic(SchemaOrgService):
    """
    Implementation of Service which is intended to parse RateYourMusic.
    Yes, because I can.
    """
    pass



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of FreeMusicArchive.
http://freemusicarchive.com
"""


from . import Service, HTMLMixin

from six import text_type


##############################################################################


class FreeMusicArchive(HTMLMixin, Service):
    """
    Implementation of Service which is intended to parse FreeMusicArchive.
    """

    def fetch_tracks(self, soup):
        return soup.select("#content div.playlist div.play-item")

    def fetch_name(self, soup, container):
        return container.find("span", class_="playtxt").a

    def fetch_track_length(self, soup, container):
        length = container.find("span", class_="playtxt")
        length.a.decompose()
        length.b.decompose()
        return text_type(length).strip("() ")



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of MusicBrainz.
http://musicbrainz.org
"""


from . import Service, XMLMixin

from six import text_type
from six.moves.urllib.parse import urlparse
from tornado.httpclient import HTTPRequest
from tornado.httputil import url_concat


##############################################################################


class MusicBrainz(XMLMixin, Service):
    """
    Implementation of Service which is intended to parse MusicBrainz.
    """

    def generate_request(self):
        url = urlparse(self.url).path.rstrip("/").rpartition("/")[-1]
        url = "http://musicbrainz.org/ws/2/release/" + url
        url = url_concat(url, dict(inc="recordings"))
        return HTTPRequest(url, use_gzip=True, user_agent=self.USER_AGENT)

    def fetch_tracks(self, soup):
        return soup.find_all("track")

    def fetch_name(self, soup, container):
        return container.find("recording").title

    def fetch_track_length(self, soup, container):
        time = container.find("length")
        if not time:
            return ""
        # text_type is a must here because we have to convert Water
        # to text first.
        return self.second_to_timestamp(int(text_type(time)) // 1000)



# -*- coding: utf-8 -*-
"""
This module contains capitalization definitions for RYM in Python 2.
"""


from .. import NLTK_PATH
from ..utils import text_type

from itertools import chain

from nltk import word_tokenize, wordpunct_tokenize, sent_tokenize, pos_tag, \
    data as NLTK_DATA
from nltk.corpus import stopwords


###############################################################################


# This small hack (?) helps NLTK to find its files.
NLTK_DATA.path[0:0] = [NLTK_PATH]


###############################################################################


PRECALCULATED_LANGSETS = {}
for _language in stopwords.fileids():
    stopwords_set = set(text_type(wrd) for wrd in stopwords.words(_language))
    stopwords_set = (wordpunct_tokenize(word) for word in stopwords_set)
    PRECALCULATED_LANGSETS[_language] = set(chain.from_iterable(stopwords_set))

CYRILLIC_ALPHABET = text_type("ЙЦУКЕНГШЩЗХЪФЫВАПРОЛДЖЭЯЧСМИТЬЬБЮ")
CYRILLIC_ALPHABET = frozenset(CYRILLIC_ALPHABET + CYRILLIC_ALPHABET.lower())


###############################################################################


class Capitalization(object):
    """
    Base capitalization framework class. Inherit to specify the logic.
    """

    def capitalize_sentence(self, sentence):
        """
        Capitalizes particular sentence.
        """
        return " ".join(chunk.capitalize() for chunk in sentence.split())

    def capitalize(self, text):
        """
        Capitalizes whole text.
        """
        text = text_type(text)
        return " ".join(
            self.capitalize_sentence(sent) for sent in sent_tokenize(text)
        )


class EnglishCapitalization(Capitalization):
    """
    Implementation of capitalization for English language.
    """

    # POS tags detected by NLTK for the words have to be lowercased.
    # Checkout a list here:
    # http://ling.upenn.edu/courses/Fall_2003/ling001/penn_treebank_pos.html
    POS_TAG_LOWER_CASE = {"CC", "IN", "TO"}
    WORD_LOWER_CASE = {"vs.", "v.", "etc", "the"}

    def capitalize_sentence(self, sentence):
        words = word_tokenize(sentence)
        tagged_words = pos_tag([word.lower() for word in words])

        truecased_words = []
        for word, tag in tagged_words:
            it_is_okay = (
                word == "n't" or
                tag in self.POS_TAG_LOWER_CASE or
                word in self.WORD_LOWER_CASE
            )
            if not it_is_okay:
                if "-" in word or "." in word:
                    word = word.title()
                elif word == "``":
                    word = '"'
                else:
                    word = word.capitalize()
            truecased_words.append(word)

        for idx in (0, -1):
            word = truecased_words[idx]
            if word != word.upper():
                truecased_words[idx] = word.capitalize()

        coerced_sentence = " ".join(truecased_words)

        return coerced_sentence


class WorldCapitalization(Capitalization):
    """
    Base class for capitalization for non-english languages.
    """

    def apply_specifics(self, sentence):
        """
        Applies language specifics on a sentence.
        """
        return sentence

    def capitalize_sentence(self, sentence):
        sentence = super(WorldCapitalization, self).capitalize_sentence(
            sentence
        )
        return self.apply_specifics(sentence)


class RussianCapitalization(WorldCapitalization):
    """
    Implementation of Russian capitalization.
    """

    def apply_specifics(self, sentence):
        return sentence.capitalize()


class SpanishCapitalization(WorldCapitalization):
    """
    Implementation for Spanish capitalization.
    """

    PREPOSITIONS = (
        "bajo",
        "con",
        "contra",
        "antes de",
        "cerca de",
        "delante de",
        "dentro de",
        "encima de",
        "desde",
        "durante",
        "enfrente de",
        "fuera de",
        "entre",
        "hacia",
        "hasta",
        "para",
        "por",
        "sin",
        "sobre",
        "tras",
        "según",
        "detrás de",
        "después de"
        "en",
        "de",
    )
    PLACE_IDS = (
        "río",
        "lago",
        "monte"
    )
    LANGUAGES = (
        "albano",
        "alemán",
        "andorrano",
        "austriaco",
        "belga",
        "bielorruso",
        "bosnio",
        "búlgaro",
        "checo",
        "croata",
        "danés",
        "eslovaco",
        "esloveno",
        "español",
        "estonio",
        "fines",
        "francés",
        "griego",
        "holandés",
        "húngaro",
        "inglés",
        "irlandés",
        "islandés",
        "italiano",
        "letón",
        "lituano",
        "luxemburgués",
        "macedonio",
        "montenegrino",
        "noruego",
        "polaco",
        "portugués",
        "rumano",
        "ruso",
        "servio",
        "sueco",
        "suizo",
        "ucraniano",
        "argentino",
        "boliviano",
        "brasileño",
        "canadiense",
        "chileno",
        "colombiano",
        "costarricense",
        "cubano",
        "dominicano",
        "ecuatoriano",
        "salvadoreño",
        "guatemalteco",
        "hondureño",
        "mexicano",
        "nicaragüense",
        "panameño",
        "paraguayo",
        "peruani",
        "portorriqueño",
        "estadounidense",
        "uruguayo",
        "venezolanoa",
        "argentino",
        "boliviano",
        "brasileño",
        "canadiense",
        "chileno",
        "colombiano",
        "costarricense",
        "cubano",
        "dominicano",
        "ecuatoriano",
        "salvadoreño",
        "guatemalteco",
        "hondureño",
        "mexicano",
        "nicaragüense",
        "panameño",
        "paraguayo",
        "peruani",
        "portorriqueño",
        "estadounidense",
        "uruguayo",
        "venezolanoa",
    )
    SUBSTITUTIONS = chain(PREPOSITIONS, PLACE_IDS, LANGUAGES)
    SUBSTITUTIONS = [text_type(sbst) for sbst in SUBSTITUTIONS]

    def apply_specifics(self, sentence):
        tokens = sentence.split(" ")
        first, rest = tokens[0], " ".join(tokens[1:])
        for prep in chain(self.SUBSTITUTIONS):
            rest = rest.replace(prep.capitalize(), prep)
        return " ".join([first, rest])


class DutchCapitalization(EnglishCapitalization):
    """
    Implementation of RYM capitalization for Dutch.
    """

    def capitalize_sentence(self, sentence):
        sentence = super(DutchCapitalization, self).capitalize_sentence(
            sentence
        )
        return sentence.replace("Ij", "IJ")


##############################################################################


def capitalize(text):
    """
    Text capitalizator for Python 2.
    """
    text = text_type(text)
    if set(text) & CYRILLIC_ALPHABET:
        language = "russian"
    else:
        words = set(wordpunct_tokenize(text_type(text).lower()))
        language = max(
            stopwords.fileids(),
            key=lambda lang: len(words & PRECALCULATED_LANGSETS[lang])
        )

    class_ = EnglishCapitalization
    if language == "russian":
        class_ = RussianCapitalization
    elif language == "spanish":
        class_ = SpanishCapitalization
    elif language == "dutch":
        class_ = DutchCapitalization
    return class_().capitalize(text)



# -*- coding: utf-8 -*-
"""
Core logic of RYMTracks. Actually there are some functions which uses
Tornado IO loop and invokes parsers.
"""


from .services import Service

from multiprocessing import cpu_count

from concurrent.futures import ThreadPoolExecutor
from tornado.ioloop import IOLoop
from tornado.httpclient import AsyncHTTPClient
from tornado.gen import coroutine, Return


##############################################################################


__all__ = "process_urls", "execute"


##############################################################################


try:
    import pycurl
    AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")
except ImportError:
    pass


##############################################################################


@coroutine
def process_urls(urls, worker_pool):
    """
    Processes a list of urls with given worker pool.
    """
    tasks = [Service.produce(url, worker_pool) for url in urls]
    tasks = [task.get_task() for task in tasks if task is not None]
    finished_tasks = yield tasks
    raise Return(finished_tasks)


##############################################################################


def execute(urls):
    """
    Just some function to be executed by main function.
    """
    worker_pool = ThreadPoolExecutor(cpu_count() * 2 + 1)
    return IOLoop.instance().run_sync(lambda: process_urls(urls, worker_pool))



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of Discogs.
http://discogs.com
"""


from . import SchemaOrgService


##############################################################################


class Discogs(SchemaOrgService):
    """
    Implementation of Service which is intended to parse Discogs.
    """
    pass



# -*- coding: utf-8 -*-
"""
This package contains implementation of proper RYM capitalization.

Please be noticed that only English capitalization is more or less supported
because it is too hard to recognize language of title based on short texts.
By default title is considered as English one. Russian capitalization is
supported so-so, spanish is more or less suits the needs but anyway, since
natural language processing is rather hard it is better not to expect
absolutely valid results in all cases here.

We cannot trust other titles because there is a common practice that artists
put names in all possible languages.

So do not forget to fix some minor problems by yourself. This module just
tries to reduce your pain.
"""


from re import compile as regex_compile, IGNORECASE as re_IGNORECASE, \
    VERBOSE as re_VERBOSE, UNICODE as re_UNICODE
from warnings import warn

from six import PY2, text_type


###############################################################################


__all__ = "capitalize",


###############################################################################


if not PY2:
    from .py3 import capitalize as specific_capitalize
else:
    try:
        from .py2 import capitalize as specific_capitalize
    except LookupError:
        warn(
            "NLTK data cannot be found. "
            "Please run rymtracks --update-languages to update it. "
            "RYMTracks won't do any capitalization without NLTK data"
        )
        from .py3 import capitalize as specific_capitalize


###############################################################################


FIX_PUNCTUATION = regex_compile(r"\s+(?=[\.,!?:;])")
SEPARATORS = regex_compile(r"([\[\]\(\)'\"\{\}\.,?!:;])")
FIX_LEFT_QUOTES = regex_compile(r"(?<=['\"])\s+")
FIX_RIGHT_QUOTES = regex_compile(r"(?<!\w|\d)\s+(?=['\"])", re_UNICODE)
FIX_LEFT_BRACES = regex_compile(r"(?<=[\[\(\}])\s+")
FIX_RIGHT_BRACES = regex_compile(r"\s+(?=[\]\}\)])")
FIX_SPACES = regex_compile(r"\s{2,}")
FIX_QUOTES = regex_compile(r"'{2,}")
FIX_SHORT_FORMS = regex_compile(r"\s+'(re|s|t)\b", re_IGNORECASE)
ROMAN_NUMERALS = regex_compile(
    r"""
        \b
            M{0,4}
            (CM|CD|D?C{0,3})
            (XC|XL|L?X{0,3})
            (IX|IV|V?I{0,3})
        \b
    """,
    re_VERBOSE | re_IGNORECASE
)


###############################################################################


def fix_roman_numeral(matcher):
    """
    Uppercases roman numerals.
    """
    return matcher.group(0).upper()


def fix_short_form(matcher):
    return "'" + matcher.group(1).lower()


def capitalize(text):
    """
    Text capitalizator. Give it a text, it will try to capitalize it according
    to the rules of the language and returns somewhat properly capitalized
    result.
    """
    text = text_type(text)
    text = SEPARATORS.sub(r" \1 ", text)

    text = " / ".join(specific_capitalize(part) for part in text.split("/"))

    text = FIX_LEFT_QUOTES.sub("", text)
    text = FIX_RIGHT_QUOTES.sub("", text)
    text = FIX_LEFT_BRACES.sub("", text)
    text = FIX_RIGHT_BRACES.sub("", text)
    text = FIX_SPACES.sub("", text)
    text = FIX_QUOTES.sub('"', text)
    text = FIX_PUNCTUATION.sub("", text)

    text = FIX_SHORT_FORMS.sub(fix_short_form, text)
    text = ROMAN_NUMERALS.sub(fix_roman_numeral, text)
    text = text.replace(" MIX", " Mix")

    return text.strip()



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of AllMusic Guide service.
http://allmusic.com
"""


from . import SchemaOrgService


##############################################################################


class AllMusic(SchemaOrgService):
    """
    Implementation of Service which is intended to parse AMG.
    """
    pass



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of iTunes.
http://itunes.apple.com
"""


from . import Service, HTMLMixin


##############################################################################


class ITunes(HTMLMixin, Service):
    """
    Implementation of Service which is intended to parse iTunes.
    """

    def fetch_tracks(self, soup):
        tracks = soup.select(
            "div.track-list .tracklist-content-box > .tracklist-table"
        )
        return tracks[0].find_all("tr", class_="song")

    def fetch_name(self, soup, container):
        return container.find("td", class_="name").find("span", class_="text")

    def fetch_track_length(self, soup, container):
        return container.find("td", class_="time").find("span", class_="text")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
RYMTracks scraps given URLs and presents tracklists into copypasteable form
for RateYourMusic.com.

Usage:
    rymtracks <url>...
    rymtracks -f <filename>
    rymtracks -l
    rymtracks --update-languages
    rymtracks (-h | --help)
    rymtracks --version

Options:
    -h --help           Show this screen.
    --version           Show version.
    --update-languages  Installs languages datafiles for proper RYM
                        capitalization
    -f                  Filename with urls.
    -l                  Show parseable network locations.
"""


from os import makedirs
from os.path import exists, expanduser, join as path_join
from shutil import rmtree
from sys import exit as sysexit

from docopt import docopt
from six import PY2, text_type, print_


###############################################################################


__version__ = 0, 1, 3, "fix1"
__all__ = 'main',


###############################################################################


HOME_PATH = path_join(expanduser("~"), ".rymtracks")
NLTK_PATH = path_join(HOME_PATH, "nltk")


###############################################################################


def print_service_locations():
    """
    Just prints recognizible network locations.
    """
    from .services import Service

    for key in Service.network_locations():
        print_(key)


def update_languages():
    """
    Updates NLTK data in your local home. Does nothing on Python 3 because
    NLTK does not support it yet.
    """
    if not PY2:
        return

    from nltk import download

    if not exists(HOME_PATH):
        makedirs(HOME_PATH)
    rmtree(NLTK_PATH, True)
    makedirs(NLTK_PATH)

    for data in ("stopwords", "punkt", "maxent_treebank_pos_tagger"):
        download(data, download_dir=NLTK_PATH)


def main():
    """
    Main function. rymtracks script executes this.
    """
    opts = docopt(
        __doc__,
        version="RYMTracks {}".format(
            ".".join(str(num) for num in __version__)
        )
    )

    urls = [text_type(url) for url in opts["<url>"]]
    if opts["<filename>"]:
        with open(opts["<filename>"], "r") as res:
            urls.extend(text_type(url) for url in res.readlines())
    if opts["-l"]:
        return print_service_locations()
    if opts["--update-languages"]:
        return update_languages()

    from .core import execute
    from .formatters import console

    console(execute(urls))


##############################################################################


if __name__ == "__main__":
    sysexit(main())



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of ProgArchives.
http://progarchives.com
"""


from . import Service, HTMLMixin
from ..capitalization import capitalize

from re import compile as regex_compile, VERBOSE as regex_VERBOSE

from six import text_type


##############################################################################


class ProgArchives(HTMLMixin, Service):
    """
    Implementation of Service which is intended to parse ProgArchives.
    """

    TRACK_REGEXP = regex_compile(
        r"""
            ^
            (?P<number>\d+)\.\s*
            (?P<title>.*?)\s*
            (?:
                \(
                    (?P<length>(?:\d+:)*\d+)
                \)
            )?\s*
            $
        """,
        regex_VERBOSE
    )

    def parse(self, response):
        converted_response = self.convert_response(response)

        tracks = converted_response.table.find_all("td")
        tracks = tracks[-1].p
        tracks = [
            text_type(chunk).strip() for chunk in tracks.stripped_strings
        ]

        extracted_data = []
        for track in tracks:
            matcher = self.TRACK_REGEXP.search(track)
            if not matcher:
                continue
            matches = matcher.groupdict()
            title = capitalize(matches["title"])
            length = self.normalize_track_length(matches["length"] or "")
            extracted_data.append((title, length))

        if not extracted_data:
            raise Exception("Empty list")
        return tuple(extracted_data)



# -*- coding: utf-8 -*-
"""
Services package. This module contains definitions of useful mixins and
main service class.
"""


from __future__ import division

from ..capitalization import capitalize

from collections import namedtuple

from bs4 import BeautifulSoup
from isodate import parse_duration, ISO8601Error
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.gen import coroutine, Return
from six import PY2, u, text_type, string_types, \
    callable as six_callable, Iterator
from six.moves.urllib.parse import urlparse

try:
    # http://simplejson.readthedocs.org. Since it is wide-spread I am sure
    # you have it
    from simplejson import loads as json_loads
except ImportError:
    from json import loads as json_loads


##############################################################################


__all__ = "Service", "HTMLMixin", "JSONMixin", "XMLMixin"

# Response structure for parsed results. exception is None if nothing happened
ParserResponse = namedtuple("ParserResponse", ["url", "data", "exception"])


##############################################################################


class Water(Iterator):
    """
    Just simple wrapper around BeautifulSoup to avoid "if"-mess.
    """

    __slots__ = "_soup",

    # ------------------------------------------------------------------------

    def __init__(self, soup):
        self._soup = soup

    def __nonzero__(self):
        return bool(self._soup) if self._soup else False

    def __instancecheck__(self, instance):
        return isinstance(instance, getattr(self._soup, "__class__", None))

    def __len__(self):
        return len(self._soup) if self._soup else 0

    def __getitem__(self, key):
        return Water(self._soup[key]) if self._soup else self

    def __setitem__(self, key, value):
        if self._soup:
            self._soup[key] = value

    def __delitem__(self, key):
        if self._soup:
            del self._soup[key]

    def __iter__(self):
        return Water(iter(self._soup)) if self._soup else iter([])

    def __contains__(self, key):
        return key in self._soup if self._soup else False

    def __getslice__(self, i, j):
        return Water(self._soup[i:j]) if self._soup else Water([])

    def __getattr__(self, item):
        if hasattr(self._soup, item):
            return Water(getattr(self._soup, item))
        return self

    def __call__(self, *args, **kwargs):
        if six_callable(self._soup):
            return Water(self._soup(*args, **kwargs))
        return self

    def __str__(self):
        if self._soup:
            text = self._soup
            if not isinstance(self._soup, string_types):
                text = self._soup.get_text()
            return text_type(text.strip())
        else:
            return text_type("")

    if PY2:
        def __unicode__(self):
            return self.__str__()

    def __repr__(self):
        return repr(self._soup)

    def __next__(self):
        return Water(next(self._soup))


##############################################################################


class HTMLMixin(object):
    """
    Mixin which asserts that response contains HTML and converts it into
    Beautiful Soup instance.
    """

    @staticmethod
    def convert_response(response):
        """
        Converter of response into Beautiful Soup instance.
        """
        return Water(BeautifulSoup(response.buffer, "html"))


class JSONMixin(object):
    """
    Mixin which asserts that response contains JSON and parses it.
    """

    @staticmethod
    def convert_response(response):
        """
        Converts response into Python objects.
        """
        raw_string = u(response.body)
        if not PY2:
            raw_string = raw_string.decode("utf-8")
        return json_loads(raw_string)


class XMLMixin(object):
    """
    Mixin which asserts that response contains XML and converts it into
    Beautiful Soup instance.
    """

    @staticmethod
    def convert_response(response):
        """
        Converter of response into Beautiful Soup instance.
        """
        return Water(BeautifulSoup(response.buffer, "xml"))


##############################################################################


class ServiceFactoryMixin(object):
    """
    Service mixin provides interface for registering Service instances,
    factory method to produce instances according to URL etc.
    """

    # A collection of parser classes binded to network locations.
    _PARSERS = []

    # ------------------------------------------------------------------------

    @classmethod
    def register(cls, class_, *netlocs):
        """
        Registers parser class to the list of network locations. Please
        remember that lookup is done by longest name so if you want
        """
        cls._PARSERS += [(loc, class_) for loc in netlocs]
        cls._PARSERS.sort(key=lambda el: len(el[0]), reverse=True)

    @classmethod
    def produce(cls, url, worker_pool):
        """
        Factory method to return appropriate Service instance for given URL.
        """
        parsed_url = urlparse(url)
        for loc, class_ in cls._PARSERS:
            if parsed_url.netloc.endswith(loc):
                return class_(url, worker_pool)

    @classmethod
    def network_locations(cls):
        """
        Returns the list of parseable network locations.
        """
        return sorted(loc for loc, class_ in cls._PARSERS)


class Service(ServiceFactoryMixin):
    """
    Main service class contains almost all logic of service handling.

    It fetches data by URL, parses it and generates ParserResponse result.

    Main method to invoke is get_task which is suitable to run in Tornado
    IOLoop. It is actually coroutine which raises result. It invokes internal
    methods in following order:

        1. generate_request
        2. parse
        3. convert_response
        4. fetch_tracks
        5. fetch_name
        6. fetch_time

    Please check documentation on methods to realise what do you need to
    override to reach the goal.
    """

    USER_AGENT = u(
        "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:29.0) "
        "Gecko/20100101 Firefox/29.0"
    )

    # ------------------------------------------------------------------------

    @staticmethod
    def normalize_track_length(length):
        """
        Removes padded zeroes from time. In other words it converts length
        into RYM-approved form: 00:13 -> 0:13, 00:03 -> 0:03
        """
        if not length:
            return ""
        chunks = [int(chunk) for chunk in length.split(":")]
        while chunks[0] == 0:
            chunks = chunks[1:]
        if len(chunks) == 1:
            chunks = [0] + chunks
        chunks = [text_type(chunk) for chunk in chunks]
        chunks[1:] = [chunk.zfill(2) for chunk in chunks[1:]]
        return ":".join(chunks)

    @staticmethod
    def second_to_timestamp(seconds):
        """
        Converts seconds into text timestamp in RYM-approved format.
        """
        parts = []
        seconds = abs(seconds)
        while seconds:
            parts.append(seconds % 60)
            seconds //= 60

        if not parts:
            return ""
        if len(parts) == 1:
            parts[0:0] = [0]

        parts = [text_type(part) for part in reversed(parts)]
        parts[1:] = [part.zfill(2) for part in parts[1:]]
        return ":".join(parts)

    # ------------------------------------------------------------------------

    def __init__(self, url, worker_pool):
        """
        Constructor.

        url is url to fetch and worker_pool is  concurrent.futures-compatible
        worker pool to process fetched data.
        """
        self.url = url
        self.worker_pool = worker_pool

    # ------------------------------------------------------------------------

    @coroutine
    def get_task(self):
        """
        Main interface method user has to invoke. It returns Tornado future
        which you have to yield to Tornado IOLoop to get the ParserResponse
        result.
        """
        raw_response = yield AsyncHTTPClient().fetch(self.generate_request())
        future = self.worker_pool.submit(self.parse, raw_response)
        if future.exception():
            response = ParserResponse(self.url, [], future.exception())
        else:
            response = ParserResponse(self.url, future.result(), None)
        raise Return(response)

    def generate_request(self):
        """
        Generates request for Tornado AsyncHTTPClient. We have to use this
        method because you might want to rewrite URL (check MusicBrainz)
        and add custom User-Agent (hello Discogs!).
        """
        return HTTPRequest(
            self.url, use_gzip=True, user_agent=self.USER_AGENT
        )

    def parse(self, response):
        """
        The most common page parsing operation. First it converts result
        into parsed object instance and after that travers it. Actually
        it has to return some properly sorted tuple with pairs of name
        and track length.
        """
        converted_response = self.convert_response(response)
        tracks = self.fetch_tracks(converted_response)
        if not tracks:
            raise Exception("Empty list")
        extracted_data = []
        for container in tracks:
            title = self.fetch_name(converted_response, container)
            title = capitalize(title)
            time = self.fetch_track_length(converted_response, container)
            time = self.normalize_track_length(text_type(time))
            extracted_data.append((title, time))
        return tuple(extracted_data)

    # ------------------------------------------------------------------------

    @staticmethod
    def convert_response(response):
        """
        Converts raw Tornado HTTPResponse into something meaningful.
        """
        return u(response.body)

    def fetch_tracks(self, response):
        """
        Fetches track containers from converted response.
        """
        raise NotImplementedError("Not implemented in base class")

    def fetch_name(self, response, container):
        """
        Fetches track name from track container.
        """
        raise NotImplementedError("Not implemented in base class")

    def fetch_track_length(self, response, container):
        """
        Fetches track length from track container.
        """
        raise NotImplementedError("Not implemented in base class")


class SchemaOrgService(HTMLMixin, Service):
    """
    Base service implementation to parse services which supports
    MusicRecording schema.

    http://schema.org/MusicRecording
    """

    def fetch_tracks(self, soup):
        return soup.find_all(
            itemtype="http://schema.org/MusicRecording",
            itemprop="track"
        )

    def fetch_name(self, soup, container):
        return container.find(itemprop="name")

    def fetch_track_length(self, soup, container):
        iso = text_type(container.find(itemprop="duration")["content"])
        try:
            return self.second_to_timestamp(parse_duration(iso).seconds)
        except ISO8601Error:
            return ""


##############################################################################


# Registering services
from .allmusic import AllMusic
from .amazon import Amazon
from .archiveorg import ArchiveOrg
from .bandcamp import BandCamp
from .beatport import BeatPort
from .boomkat import Boomkat
from .discogs import Discogs
from .freemusicarchive import FreeMusicArchive
from .itunes import ITunes
from .jamendo import Jamendo
from .junodownload import JunoDownload
from .lastfm import LastFM
from .musicbrainz import MusicBrainz
from .progarchives import ProgArchives
from .rateyourmusic import RateYourMusic
from .sevendigital import SevenDigital

Service.register(AllMusic, "allmusic.com")
Service.register(Amazon, "amazon.com", "amazon.co.uk", "amazon.co.jp")
Service.register(ArchiveOrg, "archive.org")
Service.register(BandCamp, "bandcamp.com")
Service.register(BeatPort, "beatport.com")
Service.register(Boomkat, "boomkat.com")
Service.register(Discogs, "discogs.com")
Service.register(
    FreeMusicArchive,
    "freemusicarchive.org", "freemusicarchive.com"
)
Service.register(ITunes, "itunes.apple.com")
Service.register(Jamendo, "jamendo.com")
Service.register(JunoDownload, "junodownload.com")
Service.register(
    LastFM,
    "last.fm", "lastfm.ru", "lastfm.de", "lastfm.es", "lastfm.fr",
    "lastfm.it", "lastfm.jp", "lastfm.pl", "lasfm.com.br", "lastfm.se",
    "lastfm.com.tr", "cn.last.fm"
)
Service.register(MusicBrainz, "musicbrainz.org")
Service.register(ProgArchives, "progarchives.com")
Service.register(RateYourMusic, "rateyourmusic.com")
Service.register(SevenDigital, "7digital.com")



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of BandCamp.
http://bandcamp.com
"""


from . import SchemaOrgService


##############################################################################


class BandCamp(SchemaOrgService):
    """
    Implementation of Service which is intended to parse BandCamp.
    """

    def fetch_tracks(self, soup):
        return soup.find_all(
            itemtype="http://www.schema.org/MusicRecording",
            itemprop="tracks"
        )

    def fetch_track_length(self, soup, container):
        return container.find("span", class_="time")



# -*- coding: utf-8 -*-
"""
Just a collections ot utilites used here and there.
"""


from six import PY2, text_type as text_type_


###############################################################################


def text_type(text):
    """
    text_type and u from six won't give you good results anytime if you know
    that charset is UTF-8.
    """
    if PY2 and isinstance(text, str):
        return text.decode("utf-8")
    return text_type_(text)



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of Amazon MP3 store.
http://amazon.com
"""


from . import Service, HTMLMixin

from re import compile as regex_compile


##############################################################################


class Amazon(HTMLMixin, Service):
    """
    Implementation of Service which is intended to parse Amazon MP3 Store.
    """

    # This regexp is intended to remove leading track numbers from titles.
    LEADING_NUMBER = regex_compile(r"^\d+\.\s*")

    # ------------------------------------------------------------------------

    def fetch_tracks(self, soup):
        table = soup.select("#albumTrackList > table")[1].table
        return table.find_all("tr")[1:-1]

    def fetch_name(self, soup, container):
        name = container.find("td", class_="titleCol")
        return self.LEADING_NUMBER.sub("", str(name))

    def fetch_track_length(self, soup, container):
        return container.find("td", class_="runtimeCol")



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of Archive.org.
http://archive.org
"""


from . import Service, JSONMixin
from ..capitalization import capitalize

from six import text_type, iteritems
from tornado.httpclient import HTTPRequest


##############################################################################


class ArchiveOrg(JSONMixin, Service):
    """
    Implementation of Service which is intended to parse Archive.org.
    """

    def generate_request(self):
        resource = self.url.rstrip("/").rpartition("/")[-1]
        return HTTPRequest(
            "http://archive.org/metadata/" + resource + "/files/",
            use_gzip=True,
            headers=dict(Accept="application/json")
        )

    def parse(self, response):
        converted_response = self.convert_response(response)

        tracks = {}
        required_fields = ("title", "track", "album")
        for file_ in converted_response["result"]:
            if file_.get("source") != "original":
                continue
            if not all(field in file_ for field in required_fields):
                continue

            track = int(file_["track"])
            title = capitalize(file_["title"])
            length = text_type(file_.get("length", ""))
            if length and ":" not in length:
                length = int(float(length))
                length = self.second_to_timestamp(length)
            length = self.normalize_track_length(length)
            tracks[track] = (title, length)

        if not tracks:
            raise Exception("Empty list")
        return tuple(data for track, data in sorted(iteritems(tracks)))



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of Boomkat.
http://boomkat.com
"""


from . import Service, HTMLMixin

from re import compile as regex_compile

from six import text_type


##############################################################################


class Boomkat(HTMLMixin, Service):
    """
    Implementation of Service which is intended to parse Boomkat.
    """

    # This regexp is intended to remove leading track numbers from titles.
    LEADING_NUMBER = regex_compile(r"^\d+\.\s*")

    def fetch_tracks(self, soup):
        return soup.select("#tracks-background-top div.tracks-listing")

    def fetch_name(self, soup, container):
        title = container.find("div", class_="track-listing-title")
        title.span.decompose()
        title = text_type(title)
        title = self.LEADING_NUMBER.sub("", title)
        title = title.lstrip(": ").rstrip()
        return title

    def fetch_track_length(self, soup, container):
        return container.find("div", class_="track-listing-duration ")



# -*- coding: utf-8 -*-
"""
This module contains capitalization definitions for RYM in Python 3.
"""


###############################################################################


def capitalize(text):
    """
    Python 3 primitive stub version of text capitalization. Actually it does
    nothing serious since NLTK is not ported to Python 3 yet.
    """
    return text


# -*- coding: utf-8 -*-
"""
Collection of formatters for RYMTracks.
"""


from six import print_, text_type


##############################################################################


# Width of terminal. Used by stdout formatter
STDOUT_WIDTH = 80


##############################################################################


def console(results):
    """
    Just prints results into your neat terminal window.
    """
    results = sorted(results, key=lambda res: (res.exception is None, res.url))

    for idx, result in enumerate(results):
        if result.exception:
            title = result.url + u" ({}) ".format(text_type(result.exception))
            print_(title.ljust(STDOUT_WIDTH, "="))
            continue
        if len(results) > 1:
            title = result.url + u" "
            print_(title.ljust(STDOUT_WIDTH, "="))
        for track_idx, (title, length) in enumerate(result.data, start=1):
            track_idx = str(track_idx)
            title = title.replace("|", r"\|")
            print_(track_idx + "|" + title + "|" + length)
        if idx < len(results) - 1:
            print_("")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Setup script for RYMTracks.
"""


from distutils.command.install_data import install_data
from sys import version_info as python_version
from setuptools import setup, find_packages
from setuptools.command.install import install


##############################################################################


class UpdateNLTKData(install_data):
    """
    Updater of NLTK data
    """

    def run(self):
        from os import makedirs
        from os.path import expanduser, exists, join
        from shutil import rmtree
        from nltk import download

        home_directory = join(expanduser("~"), ".rymtracks")
        nltk_data_directory = join(home_directory, "nltk")
        if not exists(home_directory):
            makedirs(home_directory)
        rmtree(nltk_data_directory, True)
        makedirs(nltk_data_directory)

        for data in ("stopwords", "punkt", "maxent_treebank_pos_tagger"):
            download(data, download_dir=nltk_data_directory)


class Install(install):
    """
    Custom procedure which updates NLTK data on install
    """

    def do_egg_install(self):
        install.do_egg_install(self)
        self.run_command("update_nltk_data")


##############################################################################


REQUIREMENTS = [
    "tornado==3.2",
    "beautifulsoup4==4.3.2",
    "lxml==3.3.0",
    "isodate==0.4.9",
    "docopt==0.6.1",
    "nose==1.3.0",
    "six==1.5.2",
    "nltk==2.0.4",
    "numpy==1.8.0"
]

EXTRAS = {}


##############################################################################


if python_version < (3,):
    REQUIREMENTS.append("futures==2.1.6")
    EXTRAS["cmdclass"] = {
        "install": Install,
        "update_nltk_data": UpdateNLTKData
    }
    EXTRAS["setup_requires"] = ["nltk==2.0.4"]

with open("README.rst", "r") as resource:
    LONG_DESCRIPTION = resource.read()


##############################################################################


setup(
    name="RYMTracks",
    description="RYMTracks scraps given URLs and presents tracklists into "
                "copypasteable form for RateYourMusic.com",
    long_description=LONG_DESCRIPTION,
    version="0.1.3fix1",
    author="Sergey Arkhipov",
    license="MIT",
    author_email="serge@aerialsounds.org",
    maintainer="Sergey Arkhipov",
    maintainer_email="serge@aerialsounds.org",
    url="https://github.com/9seconds/rymtracks/",
    install_requires=REQUIREMENTS,
    tests_require=["nose==1.3.0"],
    packages=find_packages(exclude=["tests"]),
    entry_points=dict(console_scripts=["rymtracks = rymtracks:main"]),
    test_suite="nose.collector",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Topic :: Internet :: WWW/HTTP :: Site Management :: Link Checking",
        "Topic :: Utilities"
    ],
    zip_safe=False,
    **EXTRAS
)



# -*- coding: utf-8 -*-
"""
This module contains Service implementation of JunoDownload.
http://junodownload.com
"""


from . import SchemaOrgService


##############################################################################


class JunoDownload(SchemaOrgService):
    """
    Implementation of Service which is intended to parse JunoDownload.
    """

    def fetch_tracks(self, soup):
        return soup.find_all(
            itemtype="http://schema.org/MusicRecording",
            itemprop="tracks"
        )

    def fetch_track_length(self, soup, container):
        return container.find(
            "div",
            class_="product_tracklist_heading_records_length"
        )


# -*- coding: utf-8 -*-
"""
This module contains Service implementation of Jamendo.
http://jamendo.com
"""


from . import SchemaOrgService


##############################################################################


class Jamendo(SchemaOrgService):
    """
    Implementation of Service which is intended to parse Jamendo.
    """
    pass
